import dataclasses
import logging
from typing import Dict
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Technique:
    """Technique defined in an Ontology"""

    iri: str  # Internationalized Resource Identifier
    names: Tuple[str]  # Human readable name (first is the perferred one)
    description: str  # Human readable description
    ontology_version: str  # Ontology version extracted from OWL
    versioned_iri: str  # Versioned IRI derived from stable IRI and ontology version

    @property
    def primary_name(self) -> str:
        return self.names[0]


BLISS_SCANINFO_CATEGORY = ""
_NEXUS_IDENTIFIER_PREFIX = "identifier_technique_"


@dataclasses.dataclass
class TechniqueMetadata:
    """Set of techniques with associated metadata for file (BLISS scan info)
    and data portal (ICAT dataset metafata)."""

    techniques: Set[Technique]

    def get_scan_metadata(
        self,
    ) -> Optional[Dict[str, Union[List[str], Dict[str, str]]]]:
        if self.techniques:
            return self._get_nxentry_children()
        return None

    def get_scan_info(self) -> Dict[str, Union[List[str], Dict[str, str]]]:
        if not self.techniques:
            return dict()
        return {
            BLISS_SCANINFO_CATEGORY: self._get_nxentry_children(),
            "scan_meta_categories": [BLISS_SCANINFO_CATEGORY],
        }

    def fill_scan_info(self, scan_info: MutableMapping) -> None:
        if not self.techniques:
            return
        scan_meta_categories = scan_info.setdefault("scan_meta_categories", list())

        if BLISS_SCANINFO_CATEGORY not in scan_meta_categories:
            scan_meta_categories.append(BLISS_SCANINFO_CATEGORY)

        nxentry_children = scan_info.get(BLISS_SCANINFO_CATEGORY)
        if nxentry_children is None:
            nxentry_children = scan_info[BLISS_SCANINFO_CATEGORY] = dict()

        self._fill_nxentry_children(nxentry_children)

    def _get_nxentry_children(self) -> Dict[str, str]:
        nxentry_children = dict()
        sorted_techniques = self._get_sorted_techniques()

        for i, technique in enumerate(sorted_techniques, 1):
            key = f"{_NEXUS_IDENTIFIER_PREFIX}{i}"
            nxentry_children[key] = technique.versioned_iri
            nxentry_children[f"{key}@type"] = "W3ID"

        return nxentry_children

    def _fill_nxentry_children(self, nxentry_children: MutableMapping) -> None:
        sorted_techniques = self._get_sorted_techniques()

        existing = {
            key: value
            for key, value in nxentry_children.items()
            if key.startswith(_NEXUS_IDENTIFIER_PREFIX) and not key.endswith("@type")
        }

        existing_iris = set(existing.values())

        used_indices = {int(key.split("_")[-1]) for key in existing}

        next_index = max(used_indices, default=0) + 1

        for technique in sorted_techniques:
            if technique.iri in existing_iris:
                continue

            key = f"{_NEXUS_IDENTIFIER_PREFIX}{next_index}"
            nxentry_children[key] = technique.versioned_iri
            nxentry_children[f"{key}@type"] = "W3ID"

            next_index += 1

    def _get_sorted_techniques(self) -> List[Technique]:
        return sorted(self.techniques, key=lambda technique: technique.primary_name)

    def fill_dataset_metadata(self, dataset: MutableMapping) -> None:
        if not self.techniques:
            return

        # The 'dataset' object from Bliss handles mutable mappings by only using
        # __getitem__ and __setitem__
        # https://gitlab.esrf.fr/bliss/bliss/-/blob/master/bliss/icat/policy.py

        # Get existing dataset technique names and pid's
        try:
            definitions = dataset["definition"].split(" ")
        except KeyError:
            definitions = list()
        try:
            pids = dataset["technique_pid"].split(" ")
        except KeyError:
            pids = list()

        # Add technique names and pid's
        techniques = dict(zip(pids, definitions))
        for technique in self.techniques:
            techniques[technique.iri] = technique.primary_name

        # Replace ICAT metadata associated to 'techniques'
        metadata = self._get_icat_metadata(techniques)
        for key, value in metadata.items():
            try:
                dataset[key] = value
            except KeyError:
                if key in ("technique_pid", "technique_pid_esrfet_version"):
                    _logger.warning(
                        f"Skip ICAT field '{key}' (requires pyicat-plus>=0.2)"
                    )
                    continue
                raise

    def get_dataset_metadata(self) -> Dict[str, str]:
        if not self.techniques:
            return dict()
        techniques = {
            technique.iri: technique.primary_name for technique in self.techniques
        }
        return self._get_icat_metadata(techniques)

    def _get_icat_metadata(self, techniques: Dict[str, str]) -> Dict[str, str]:
        if not techniques:
            return dict()
        iris, definitions = zip(*sorted(techniques.items(), key=lambda tpl: tpl[1]))
        metadata = {
            "technique_pid": " ".join(iris),
            "definition": " ".join(definitions),
        }
        if self.techniques:
            metadata["technique_pid_esrfet_version"] = (
                self._get_ontology_version_number()
            )
        return metadata

    def _get_ontology_version_number(self) -> Optional[str]:
        if self.techniques:
            return next(iter(self.techniques)).ontology_version.lstrip("v")
        return None



from functools import lru_cache
from typing import Generator
from typing import List
from typing import Optional
from typing import Set

from ..db import load_techniques as _load_techniques
from .types import BLISS_SCANINFO_CATEGORY  # noqa F401
from .types import Technique
from .types import TechniqueMetadata


def get_technique_metadata(*identifiers: str) -> TechniqueMetadata:
    """Returns an object that can generate several types of metadata
    associated to the provided technique names or IRIs."""
    return TechniqueMetadata(techniques=set(_iter_from_identifiers(*identifiers)))


def get_techniques(*identifiers: str) -> Set[Technique]:
    """Returns a set of techniques referenced by the provided technique names or IRIs."""
    return set(_iter_from_identifiers(*identifiers))


def get_technique(identifier: str) -> Technique:
    """Returns a technique referenced by the provided technique name or IRI."""
    return next(_iter_from_identifiers(identifier))


def get_ontology_version(metadata: Optional[TechniqueMetadata] = None) -> str:
    if metadata is None:
        technique = get_technique("XRAYS")
    else:
        technique = next(iter(metadata.techniques))
    return technique.ontology_version


def get_ontology_version_number(metadata: Optional[TechniqueMetadata] = None) -> str:
    return get_ontology_version(metadata=metadata).lstrip("v")


@lru_cache(maxsize=1)
def get_all_techniques() -> List[Technique]:
    """Returns a list of techniques."""
    return [
        Technique(
            iri=technique["iri"],
            names=tuple(technique["names"]),
            description=technique["description"],
            ontology_version=technique["ontology_version"],
            versioned_iri=technique["versioned_iri"],
        )
        for ontology_name in ["ESRFET"]
        for technique in _load_techniques(ontology_name)
    ]


def _iter_from_identifiers(*identifiers: str) -> Generator[Technique, None, None]:
    techniques = get_all_techniques()
    for identifier in sorted(set(identifiers)):
        identifier_lower = identifier.lower()
        for technique in techniques:
            technique_names = set(map(str.lower, technique.names))
            if (
                identifier_lower in technique_names
                or identifier == technique.iri
                or identifier == technique.versioned_iri
            ):
                yield technique
                break
        else:
            raise KeyError(f"'{identifier}' is not a known technique name or IRI.")



import re

from ..technique import get_all_techniques
from ..technique import get_ontology_version
from ..technique import get_ontology_version_number
from ..technique import get_technique
from ..technique import get_techniques


def test_get_all_technique():
    assert get_all_techniques()


def test_get_techniques():
    all_techniques = set(get_all_techniques())
    subset = get_techniques("XRF", "XRD")
    assert subset
    assert subset < all_techniques


def test_get_technique():
    all_techniques = set(get_all_techniques())
    technique = get_technique("XRF")
    assert technique in all_techniques


def test_get_technique_by_iri():
    all_techniques = set(get_all_techniques())
    technique = get_technique("https://w3id.org/PaN/ESRFET#XRF")
    assert technique in all_techniques


def test_get_technique_by_versioned_iri():
    all_techniques = set(get_all_techniques())
    version = get_ontology_version()
    technique = get_technique(f"https://w3id.org/PaN/ESRFET/{version}/#XRF")
    assert technique in all_techniques


def test_ontology_version():
    version = get_ontology_version()
    pattern = r"^v\d+\.\d+\.\d+$"
    assert re.match(pattern, version)


def test_ontology_version_number():
    version_number = get_ontology_version_number()
    pattern = r"^\d+\.\d+\.\d+$"
    assert re.match(pattern, version_number)



import pytest

from ..technique import BLISS_SCANINFO_CATEGORY
from ..technique import get_ontology_version
from ..technique import get_ontology_version_number
from ..technique import get_technique_metadata


def test_get_dataset_metadata():
    metadata = get_technique_metadata("XAS")
    ontology_version = get_ontology_version_number(metadata)
    dataset_metadata = {
        "definition": "XAS",
        "technique_pid": "https://w3id.org/PaN/ESRFET#XAS",
        "technique_pid_esrfet_version": ontology_version,
    }

    assert metadata.get_dataset_metadata() == dataset_metadata


def test_fill_dataset_metadata():
    metadata = get_technique_metadata("XAS")
    ontology_version = get_ontology_version_number(metadata)
    dataset_metadata = {
        "definition": "XAS",
        "technique_pid": "https://w3id.org/PaN/ESRFET#XAS",
        "technique_pid_esrfet_version": ontology_version,
    }

    dataset = {}
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = dict(dataset_metadata)
    metadata.fill_dataset_metadata(dataset_metadata)
    assert dataset == dataset_metadata


def test_get_scan_info():
    metadata = get_technique_metadata("XAS")
    ontology_version = get_ontology_version(metadata)
    scan_info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_1@type": "W3ID",
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info[BLISS_SCANINFO_CATEGORY]


def test_fill_scan_info():
    metadata = get_technique_metadata("XAS")
    ontology_version = get_ontology_version(metadata)
    scan_info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_1@type": "W3ID",
        },
    }
    info = {}
    metadata.fill_scan_info(info)
    assert info == scan_info

    scan_info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY, "technique"],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XRF",
            "identifier_technique_1@type": "W3ID",
            "identifier_technique_2": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_2@type": "W3ID",
        },
    }
    info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY, "technique"],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XRF",
            "identifier_technique_1@type": "W3ID",
        },
    }
    metadata.fill_scan_info(info)
    assert info == scan_info


def test_wrong_technique_metadata():
    with pytest.raises(KeyError, match="'WRONG' is not a known technique name or IRI"):
        get_technique_metadata("XAS", "WRONG")


def test_empty_technique_metadata():
    metadata = get_technique_metadata()
    assert metadata.get_dataset_metadata() == dict()
    assert metadata.get_scan_info() == dict()
    assert metadata.get_scan_metadata() is None


def test_get_metadata_from_iri():
    metadata = get_technique_metadata("https://w3id.org/PaN/ESRFET#XAS")
    dataset_metadata = metadata.get_dataset_metadata()

    assert dataset_metadata["definition"] == "XAS"
    assert dataset_metadata["technique_pid"] == "https://w3id.org/PaN/ESRFET#XAS"


def test_get_metadata_from_versioned_iri():
    version = get_ontology_version()
    metadata = get_technique_metadata(f"https://w3id.org/PaN/ESRFET/{version}/#XAS")
    dataset_metadata = metadata.get_dataset_metadata()

    assert dataset_metadata["definition"] == "XAS"
    assert dataset_metadata["technique_pid"] == "https://w3id.org/PaN/ESRFET#XAS"


def test_get_metadata_from_name():
    metadata = get_technique_metadata("xas")
    dataset_metadata = metadata.get_dataset_metadata()

    assert dataset_metadata["definition"] == "XAS"
    assert dataset_metadata["technique_pid"] == "https://w3id.org/PaN/ESRFET#XAS"



from ..technique import BLISS_SCANINFO_CATEGORY
from ..technique import get_ontology_version
from ..technique import get_ontology_version_number
from ..technique import get_technique_metadata


def test_get_dataset_metadata():
    metadata = get_technique_metadata("XRF", "XAS")
    ontology_version = get_ontology_version_number(metadata)
    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "https://w3id.org/PaN/ESRFET#XAS",
                "https://w3id.org/PaN/ESRFET#XRF",
            ]
        ),
        "technique_pid_esrfet_version": ontology_version,
    }

    assert metadata.get_dataset_metadata() == dataset_metadata


def test_fill_dataset_metadata():
    metadata = get_technique_metadata("XRF", "XAS")
    ontology_version = get_ontology_version_number(metadata)
    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "https://w3id.org/PaN/ESRFET#XAS",
                "https://w3id.org/PaN/ESRFET#XRF",
            ]
        ),
        "technique_pid_esrfet_version": ontology_version,
    }

    dataset = dict(dataset_metadata)
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {
        "definition": "XRF",
        "technique_pid": "https://w3id.org/PaN/ESRFET#XRF",
    }
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {}
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {
        "definition": "XRD",
        "technique_pid": "https://w3id.org/PaN/ESRFET#XRD",
        "technique_pid_esrfet_version": ontology_version,
    }
    dataset_metadata = {
        "definition": "XAS XRD XRF",
        "technique_pid": " ".join(
            [
                "https://w3id.org/PaN/ESRFET#XAS",
                "https://w3id.org/PaN/ESRFET#XRD",
                "https://w3id.org/PaN/ESRFET#XRF",
            ]
        ),
        "technique_pid_esrfet_version": ontology_version,
    }
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata


def test_get_scan_info():
    metadata = get_technique_metadata("XRF", "XAS")
    ontology_version = get_ontology_version(metadata)
    scan_info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_1@type": "W3ID",
            "identifier_technique_2": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XRF",
            "identifier_technique_2@type": "W3ID",
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info[BLISS_SCANINFO_CATEGORY]


def test_fill_scan_info():
    metadata = get_technique_metadata("XRF", "XAS")
    ontology_version = get_ontology_version(metadata)
    scan_info = {
        "scan_meta_categories": ["technique", BLISS_SCANINFO_CATEGORY],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_1@type": "W3ID",
            "identifier_technique_2": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XRF",
            "identifier_technique_2@type": "W3ID",
        },
    }

    info = {
        "scan_meta_categories": ["technique"],
        BLISS_SCANINFO_CATEGORY: None,
    }
    metadata.fill_scan_info(info)
    assert info == scan_info


def test_double_technique_metadata():
    metadata = get_technique_metadata("XRF", "XAS", "XRF", "XAS")
    ontology_version = get_ontology_version(metadata)
    ontology_version_number = get_ontology_version_number(metadata)
    assert len(metadata.techniques) == 2

    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "https://w3id.org/PaN/ESRFET#XAS",
                "https://w3id.org/PaN/ESRFET#XRF",
            ]
        ),
        "technique_pid_esrfet_version": ontology_version_number,
    }
    assert metadata.get_dataset_metadata() == dataset_metadata

    scan_info = {
        "scan_meta_categories": [BLISS_SCANINFO_CATEGORY],
        BLISS_SCANINFO_CATEGORY: {
            "identifier_technique_1": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XAS",
            "identifier_technique_1@type": "W3ID",
            "identifier_technique_2": f"https://w3id.org/PaN/ESRFET/{ontology_version}/#XRF",
            "identifier_technique_2@type": "W3ID",
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info[BLISS_SCANINFO_CATEGORY]


def test_multiple_identifiers_with_iri():
    metadata = get_technique_metadata(
        "XRF",
        "https://w3id.org/PaN/ESRFET#XAS",
    )

    dataset = metadata.get_dataset_metadata()

    assert set(dataset["definition"].split()) == {"XAS", "XRF"}
    assert set(dataset["technique_pid"].split()) == {
        "https://w3id.org/PaN/ESRFET#XAS",
        "https://w3id.org/PaN/ESRFET#XRF",
    }


def test_multiple_identifiers_with_versioned_iri():
    version = get_ontology_version()
    metadata = get_technique_metadata(
        "XRF",
        f"https://w3id.org/PaN/ESRFET/{version}/#XAS",
    )

    dataset = metadata.get_dataset_metadata()

    assert set(dataset["definition"].split()) == {"XAS", "XRF"}
    assert set(dataset["technique_pid"].split()) == {
        "https://w3id.org/PaN/ESRFET#XAS",
        "https://w3id.org/PaN/ESRFET#XRF",
    }
